# -*- coding: latin-1 -*- 
import array
import unicodeparser

class BinaryData16:
    def __init__(self, string):
        self.s = array.array('H',string)
    def read16(self, ofs):
        return self.s[ofs>>1]
    
    def write16(self, d,ofs):
        self.s[ofs>>1]=d      

    def read32(self, ofs):
        return self.s[ofs>>1] | (self.s[(ofs>>1)+1]<<16)

    def write32(self, d, ofs):
        self.s[ofs>>1]=(d&0xffff)
        self.s[(ofs>>1)+1]=((d>>16)&0xffff)
    def getStr(self):
        return self.s.tostring()


class BinaryData:
    def __init__(self, string):
        self.s = array.array('c',string)
    def read16(self, ofs):
        return ord(self.s[ofs]) | (ord(self.s[ofs+1])<<8)
    
    def write16(self, d,ofs):
        self.s[ofs]=chr(d&0xff)
        self.s[ofs+1]=chr((d>>8)&0xff)        

    def read32(self, ofs):
        return ord(self.s[ofs]) | (ord(self.s[ofs+1])<<8) | (ord(self.s[ofs+2])<<16) | (ord(self.s[ofs+3])<<24)

    def write32(self, d, ofs):
        self.s[ofs]=chr(d&0xff)
        self.s[ofs+1]=chr((d>>8)&0xff)
        self.s[ofs+2]=chr((d>>16)&0xff)
        self.s[ofs+3]=chr((d>>24)&0xff)
    def getStr(self):
        return self.s.tostring()
        
        

class PokeTextData(BinaryData16): #16 = mayor speed improvement
    def decrypt(self):
        self.DecyptPtrs(self.read16(0), self.read16(2), 4)
        self.ptrlist = self.CreatePtrList(self.read16(0), 4)

        self.strlist = []

        for i in range(self.read16(0)):
            ptr, chars = self.ptrlist[i]    
            self.DecyptTxt(chars, i+1, ptr)
            self.strlist.append(self.MakeString(chars, ptr))
    
    def encrypt(self):
        
        self.ptrlist = self.CreatePtrList(self.read16(0), 4)

        for i in range(self.read16(0)):
            ptr, chars = self.ptrlist[i]    
            self.DecyptTxt(chars, i+1, ptr)            

        self.DecyptPtrs(self.read16(0), self.read16(2), 4)
        
    def DecyptPtrs(self, count, key, sdidx):
        key = (key * 0x2FD) &0xffff

        for i in range(count):
            key2 = (key*(i+1)&0xffff)
            realkey = key2 | (key2<<16)
            self.write32(self.read32(sdidx)^realkey, sdidx)       
            self.write32(self.read32(sdidx+4)^realkey, sdidx+4)
            sdidx+=8
        

    def CreatePtrList(self, count, sdidx):
        ptrlist = []
        for i in range(count):
            ptrlist.extend([[self.read32(sdidx), self.read32(sdidx+4)]])
            sdidx+=8
        return ptrlist
        

        
    def DecyptTxt(self, count, id, idx):
        key = (0x91BD3*id)&0xffff

        for i in range(count):
            self.write16(self.read16(idx)^key, idx)
            key +=0x493D
            key&=0xffff
            idx+=2

    def MakeString(self, count, idx):
        string = ""
        for i in range(count):
            try:
                string = string + unicodeparser.tb[self.read16(idx)]
            except:
                if self.read16(idx)==0xfffe:
                    idx+=2
                    string = string + "\\v" + "%04X"%self.read16(idx)
                    idx+=2
                    string = string + "\\w" + "%04X"%self.read16(idx)
                elif self.read16(idx)==0xffff:
                    break                    
                else:
                    string = string + "\\x" + "%04X"%self.read16(idx)
            idx+=2
        return string
    
    def SetKey(self, key):
        self.write16(key,2)
